# -*- coding: utf-8 -*-
import requests

import datetime

from bs4 import BeautifulSoup

from bvggrabber.api import QueryApi, Departure, Response, timeformat

from bvggrabber.utils.format import dateformat, int2bin


SCHEDULED_API_ENDPOINT = 'https://mobil.bvg.de/Fahrinfo/bin/stboard.bin/dox'


class Vehicle():

    S = 64
    U = 32
    TRAM = 16
    BUS = 8
    FERRY = 4
    RB = 2
    IC = 1

    _ALL = 127


class ScheduledDepartureQueryApi(QueryApi):

    def __init__(self, station, vehicles=Vehicle._ALL, limit=5):
        """:param Vehicle vehicles: a bitmask described by :class:`Vehicle`"""
        super(ScheduledDepartureQueryApi, self).__init__()
        if isinstance(station, str):
            self.station_enc = station.encode('iso-8859-1')
        elif isinstance(station, bytes):
            self.station_enc = station
        else:
            raise ValueError("Invalid type for station")
        self.station = station
        self.vehicles = int2bin(vehicles, 7)
        self.limit = limit

    def call(self):
        params = {'input': self.station_enc,
                  'time': timeformat(datetime.datetime.now()),
                  'date': dateformat(datetime.datetime.now()),
                  'productsFilter': self.vehicles,
                  'maxJourneys': self.limit,
                  'start': 'yes'}
        response = requests.get(SCHEDULED_API_ENDPOINT, params=params)
        if response.ok:
            soup = BeautifulSoup(response.text, "html.parser")
            if soup.find('span', 'error'):
                # The station we are looking for is ambiguous or does not exist
                stations = soup.find('span', 'select').find_all('a')
                if stations:
                    # The station is ambiguous
                    stationlist = [s.text.strip() for s in stations]
                    return Response(False, stationlist)
                else:
                    # The station does not exist
                    return Response(False, [])
            else:
                # The station seems to exist
                tbody = soup.find('tbody')
                if tbody is None:
                    return Response(True, self.station, [])
                rows = tbody.find_all('tr')
                departures = []
                for row in rows:
                    tds = row.find_all('td')
                    if len(tds) != 3:
                        continue
                    dep = Departure(start=self.station,
                                    end=tds[2].text.strip(),
                                    when=tds[0].text.strip(),
                                    line=tds[1].text.strip())
                    departures.append(dep)
                return Response(True, self.station, departures)
        else:
            try:
                response.raise_for_status()
            except requests.RequestException as e:
                return Response(False, error=e)
            else:
                return Response(False,
                                error=Exception("An unknown error occured"))



#! /usr/bin/env python3
# -*- coding: utf-8 -*-

import argparse
import sys

from bvggrabber.api.actualdeparture import ActualDepartureQueryApi

from bvggrabber.api.scheduleddeparture import ScheduledDepartureQueryApi, Vehicle


if __name__ == '__main__':

    vehicle_choices = ('S', 'U', 'TRAM', 'BUS', 'FERRY', 'RB', 'IC')

    parser = argparse.ArgumentParser(
        description='Query the BVG-website for departures')
    parser.add_argument('station', type=str, help='The station to query')
    parser.add_argument('file', type=str, help='Path to file. Use - for stdout')
    parser.add_argument('--vehicle', type=str, nargs='*',
                        choices=vehicle_choices,
                        help='Vehicles which shall be queried, if non given '
                             'actualdepartue (bus) will be used')
    parser.add_argument('--limit', type=int, default=9,
                        help='Max departures to query. Default: 9')
    args = parser.parse_args()

    query = None
    res = None
    if args.vehicle:
        vehicles = 0
        bus = False
        for vehicle in args.vehicle:
            if vehicle == 'S':
                vehicles |= Vehicle.S
            elif vehicle == 'U':
                vehicles |= Vehicle.U
            elif vehicle == 'TRAM':
                vehicles |= Vehicle.TRAM
            elif vehicle == 'BUS':
                bus = True
            elif vehicle == 'FERRY':
                vehicles |= Vehicle.FERRY
            elif vehicle == 'RB':
                vehicles |= Vehicle.RB
            elif vehicle == 'IC':
                vehicles |= Vehicle.IC

        query = ScheduledDepartureQueryApi(args.station, vehicles, limit=args.limit)
        res = query.call()
        if bus:
            aquery = ActualDepartureQueryApi(args.station, limit=args.limit)
            res2 = aquery.call()
            res.merge(res2)
    else:
        query = ActualDepartureQueryApi(args.station, limit=args.limit)
        res = query.call()

    if args.file in ('stdout', '-'):
        print(res.to_json)
    else:
        with open(args.file, 'w') as f:
            print(res.to_json, file=f)


# -*- coding: utf-8 -*-
import datetime
import re

from functools import total_ordering
from math import floor

from dateutil.parser import parse

from bvggrabber.utils.format import timeformat
from bvggrabber.utils.json import ObjectJSONEncoder


def compute_remaining(start, end):
    """Compute the number of seconds between ``start`` and ``end`` and return
    the number of seconds rounded down to entire minutes. That means:

    * [0, 59] => 0
    * [60, 119] => 60
    * [120, 179] => 120
    * [-59, -1] => -60
    * [-119, -60] => -120

    :param datetime.datetime start: The start to compute the remaining number
        of minutes.
    :param datetime.datetime end: The end of the interval
    :raises: TypeError if either of ``start`` or ``end`` is not a
        ``datetime.datetime``.
    :return: The number of remaining seconds rounded to entire minutes
    :rtype: int
    """
    if not isinstance(start, datetime.datetime):
        raise TypeError("start needs to be a datetime.datetime")
    if not isinstance(end, datetime.datetime):
        raise TypeError("start needs to be a datetime.datetime")
    seconds = (end - start).total_seconds()
    return datetime.timedelta(minutes=floor(seconds / 60)).total_seconds()


class QueryApi(object):

    def __init__(self):
        pass

    def call(self):
        """Needs to be implemented by inheriting classes!"""
        raise NotImplementedError("The inheriting class needs to implement "
                                  "the call() method!")


class Response(object):

    def __init__(self, state, station=None, departures=None, error=None):
        """Creates a new response. Returned by :meth:`QueryApi.call`
        
        :param bool state: ``True`` iff the request and parsing was successful,
            ``False`` otherwise.
        :param station: If a ``list``, the station name is ambiguous. If a
            string the full qualified name of the station.
        :param departures: A list of :class:`Departure`` objects
        :param Exception error: In case an unexpected error occurred, this
            contains the original exception.

        If ``state`` is ``True``, ``station`` must be a ``str`` and
        ``departures`` must be a list of :class:`Departure`` objects. If
        ``state`` is ``False`` there must be several reasons for that:

            1. The provided station name during the :meth:`QueryApi.call`
               returned multiple possible departing stations. You have to
               specify the name in an unambiguous way.
            2. The station does not exist at all.
            3. An exception occurred during the :meth:`QueryApi.call`

        .. deprecated:: 0.1b3
           The ``state`` argument will be removed in the future and will be
           computed automatically based on ``station``, ``departures`` and
           ``error``.
        """
        self._state = state
        self._departures = [(station, departures)]
        self._error = error
        if self._error is None:
            if isinstance(station, list):
                self._state = False
                msg = ', '.join(station)
                self._error = Exception("Station is ambiguous: %s" % msg)
            elif station is None:
                self._state = False
                self._error = Exception("Station does not exist")
        elif isinstance(self._error, str):
            self._state = False
            self._error = Exception(self._error)

    def merge(self, other):
        """Checks that ``other`` is a :class:`Response` and extends
        :attr:`departures` by the departures given in ``other`` iff neither
        response object has a invalid state.
        """
        if isinstance(other, Response):
            if not other.state:
                raise ValueError("The response contains errors: " +
                                 str(other.error))
            elif not self.state:
                raise ValueError("The response contains errors: " +
                                 str(self.error))
            else:
                self.departures.extend(other.departures)
        else:
            raise TypeError("The given object is not a response object")

    @property
    def to_json(self):
        """.. deprecated:: 0.0.1
           Use :attr:`json` instead.
        """
        return ObjectJSONEncoder(ensure_ascii=False).encode(self.departures)

    @property
    def departures(self):
        """A list of 2-tuple in the form (:class:`str`, :class:`Departure`).
        The first element in the tuple defines the departing station where the
        second element holds a list of departure objects.
        """
        if self.state:
            return self._departures
        return str(self.error)

    @property
    def error(self):
        """The error that occurred during creation or ``None``"""
        return self._error

    @property
    def json(self):
        """Uses :class:`~bvggrabber.utils.json.ObjectJSONEncoder` to encode
        the :attr:`departures` to a JSON format.
        """
        if self.state:
            return ObjectJSONEncoder(ensure_ascii=False).encode(self.departures)
        return ObjectJSONEncoder(ensure_ascii=False).encode(str(self.error))

    @property
    def state(self):
        """``True`` iff the request and parsing was successful, ``False``
        otherwise.
        """
        return self._state


@total_ordering
class Departure(object):

    def __init__(self, start, end, when, line, since=None, no_add_day=False):
        """
        :param str start: The start station
        :param str end: The end station
        :param when: The leaving time of the public transport at the given
            ``start`` station. Might be an :class:`int` (timestamp), a
            :class:`datetime.datetime` instance or a :class:`str` accepted by
            ``dateutil.parse()``. If ``when`` is smaller than ``since`` and the
            difference between both times is larger than 12 hours (43200sec),
            the API will add another day unless ``no_add_day`` is ``True``.
        :param str line: The line of the public transport
        :param since: Either ``None`` or :class:`datetime.datetime`. Defines
            the temporal start for searching. ``None`` will internally be
            resolved as :meth:`datetime.datetime.now`.
        :param bool no_add_day: If true, there no additional day will be added
            if ``when`` is smaller than ``since``. Default ``False``.
        :raises: :exc:`TypeError` if ``when`` is invalid or cannot be parsed.
        """
        if since is None:
            self.now = datetime.datetime.now()
        else:
            self.now = since

        self.start = start
        self.end = end
        self.line = line
        if isinstance(when, (int, float)):
            # We assume to get a UNIX / POSIX timestamp
            self.when = datetime.datetime.fromtimestamp(when)
        elif isinstance(when, str):
            # We need to parse a string. But we need to remove trailing
            # whitespaces and *
            self.when = parse(re.sub('[\s*]$', '', when))
        elif isinstance(when, datetime.datetime):
            # Everything's fine, we can just take the parameter as is
            self.when = when
        else:
            raise TypeError("when must be a valid datetime, timestamp or "
                            "string!")
        diff = abs((self.when - self.now).total_seconds())
        if not no_add_day and self.when < self.now and diff > 43200:
            # 43200 are 12 hours in seconds So we accept a offset of 12 hours
            # that is still counted as "time gone" for the current day.
            self.when = self.when + datetime.timedelta(days=1)

    @property
    def remaining(self):
        """.. seealso:: bvggrabber.api.compute_remaining"""
        return int(compute_remaining(self.now, self.when))

    def __eq__(self, other):
        """Two departures are assumed to be equal iff their remaining time
        and their destination are equal.

        Right now we do **not** considering the start or line, since that would
        require some kind of geo location in order to define a *total order*.
        """
        return ((self.remaining, self.end.lower()) ==
                (other.remaining, other.end.lower()))

    def __lt__(self, other):
        """A departure is assumed to be less than another iff its remaining
        time is less than the remaining time of the other departure.

        Right now we do **not** considering the start, end or line, since that
        would require some kind of geo location in order to define a *total
        order*.
        """
        return (self.remaining < other.remaining)

    def __str__(self):
        return "Start: %s, End: %s, when: %s, now: %s, line: %s" % (
            self.start, self.end, timeformat(self.when),
            timeformat(self.now), self.line)

    def __repr__(self):
        return self.__str__()



#-*- coding: utf-8 -*-


def dateformat(dt):
    """Formats a :class:`datetime.datetime` object as dd.mm.yyyy

    :param datetime.datetime dt: The :class:`datetime.datetime` object to
        format
    :return: A formatted string
    :rtype: str

    """
    return dt.strftime('%d.%m.%Y')


def fullformat(dt):
    """Formats a :class:`datetime.datetime` object as YYYY-MM-DD HH:MM:SS

    :param datetime.datetime dt: The :class:`datetime.datetime` object to
        format
    :return: A formatted string
    :rtype: str

    """
    return dt.strftime('%Y-%m-%d %H:%M:%S')


def int2bin(i, length=8):
    """Returns the bit representation of the given integer with a minimum
    length of ``length``. E.g. ``int2bin(109, 7) == '1101101'`` and
    ``int2bin(109, 8) == '01101101'``.

    :param int i: The integer to format
    :param int length: The minimum length of the output string. The string is
        zero-padded on the left.
    :return: The bit representation of the given int
    :rtype: string

    """
    if not isinstance(length, int):
        raise ValueError("expected int for length")
    return ('{:0>' + str(length) + 'b}').format(i)


def timeformat(dt):
    """Formats a :class:`datetime.datetime` object as HH:MM

    :param datetime.datetime dt: The :class:`datetime.datetime` object to
        format
    :return: A formatted string
    :rtype: str

    """
    return dt.strftime('%H:%M')



# -*- coding: utf-8 -*-
import requests

from bs4 import BeautifulSoup

from bvggrabber.api import QueryApi, Departure, Response


ACTUAL_API_ENDPOINT = 'https://mobil.bvg.de/Fahrinfo/bin/stboard.bin/dox?ld=0.1&rt=0&'


class ActualDepartureQueryApi(QueryApi):

    def __init__(self, station, limit=5):
        super(ActualDepartureQueryApi, self).__init__()
        if isinstance(station, str):
            self.station_enc = station.encode('iso-8859-1')
        elif isinstance(station, bytes):
            self.station_enc = station
        else:
            raise ValueError("Invalid type for station")
        self.station = station
        self.limit = limit

    def call(self):
        params = {
            'input': self.station_enc,
            'maxJourneys': self.limit,
            'start': 'suchen',
        }
        response = requests.get(ACTUAL_API_ENDPOINT, params=params)
        if response.ok:
            soup = BeautifulSoup(response.text, "html.parser")
            if soup.find_all('form'):
                # The station we are looking for is ambiguous or does not exist
                stations = soup.find_all('option')
                if stations:
                    # The station is ambiguous
                    stationlist = [s.get('value') for s in stations]
                    return Response(False, stationlist)
                else:
                    # The station does not exist
                    return Response(False)
            else:
                # The station seems to exist
                result = soup.find('table', {'id': '',
                                           'class': 'ivu_table'})
                if result is None:
                    return Response(True, self.station, [])
                rows = result.find_all('tr')
                departures = []
                for row in rows:
                    if row.parent.name == 'tbody':
                        td = row.find_all('td')
                        if len(td) != 3:
                            continue
                        if td:
                            dep = Departure(start=self.station,
                                            end=td[2].text.strip(),
                                            when=td[0].text.strip(),
                                            line=td[1].text.strip())
                            departures.append(dep)
                return Response(True, self.station, departures)
        else:
            try:
                response.raise_for_status()
            except requests.RequestException as e:
                return Response(False, error=e)
            else:
                return Response(False,
                                error=Exception("An unknown error occured"))



# -*- coding: utf-8 -*-

import inspect

from json import JSONEncoder


def is_not_method(o):
    return not inspect.isroutine(o)


class ObjectJSONEncoder(JSONEncoder):

    def default(self, reject):
        non_methods = inspect.getmembers(reject, is_not_method)
        return {attr: value for attr, value in non_methods
                if (not attr.startswith('__')
                    and attr != 'when'
                    and attr != 'now')}



# -*- coding: utf-8 -*-
import codecs

from os import path
from setuptools import setup, find_packages

BASEDIR = path.dirname(__file__)

open = lambda filepath: codecs.open(filepath, 'r', 'utf-8')

description = open(path.join(BASEDIR, 'README.rst')).read()

setup(
    name='bvg-grabber',
    description='Querying the upcoming public transport departures in Berlin',
    long_description=description,
    version='0.2.0',
    url='https://github.com/MarkusH/bvg-grabber',
    author='Christian Struck, Markus Holtermann',
    author_email='info@markusholtermann.eu',
    license='BSD',
    packages=find_packages(exclude=['tests']),
    include_package_data=True,
    install_requires=[
        'beautifulsoup4>=4.4.1',
        'python-dateutil>=2.1',
        'requests>=1.2.3',
        'six>=1.3.0',
    ],
    test_suite="tests",
    scripts=['bvg-grabber.py'],
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
    ],
)
